from typing import Any, Dict, List

import numpy as np
import seaborn as sns
from matplotlib.axes import Axes
from pyspark.sql import DataFrame


def _format_spark_dataframe_for_visualization(
    spark_dataframe: DataFrame,
) -> List[Dict[str, int]]:
    """
    Prepare a Pandas DataFrame to be displayed as a heatmap visualizing the group and batch counts.

    Parameters
    ----------
    spark_dataframe : DataFrame
        The Spark DataFrame with `group` and `batch` columns.

    Returns
    -------
    List[Dict[str, int]]
        The Spark DataFrame contents processed and formatted as a list of dictionaries.
    """

    counts_sdf = spark_dataframe.groupBy("group", "batch").count()
    return [row.asDict() for row in counts_sdf.collect()]


def create_ingest_heatmap(
    spark_dataframe: DataFrame, title: str = "Parallel Ingest Heat Map"
) -> Axes:
    """
    Create the ingest heatmap from a list of dictionaries.
    This heatmap will display the groups on the y-axis and batches on the x-axis in sequential order.

    Parameters
    ----------
    spark_dataframe : DataFrame
        A Spark DataFrame with columns including 'group', 'batch' and 'count'
    title : str, optional
        A title for the visualization, by default "Parallel Ingest Heat Map"

    Returns
    -------
    Axes
        A Matplotlib Axes object for visualization.
    """

    data = _format_spark_dataframe_for_visualization(spark_dataframe=spark_dataframe)

    assert (
        set(data[0].keys()) == {"group", "batch", "count"}
    ), "Invalid keys detected in data. Dictionary keys must contain only 'group', 'batch' and 'count'."

    X_KEY = "batch"
    Y_KEY = "group"
    VALUE_KEY = "count"

    # Extract unique x and y values
    x_values = sorted(set(d[X_KEY] for d in data))
    y_values = sorted(set(d[Y_KEY] for d in data))

    # Create a 2D numpy array for the heatmap
    heatmap_data = np.zeros((len(y_values), len(x_values)))

    # Fill the array with values
    for item in data:
        x_idx = x_values.index(item[X_KEY])
        y_idx = y_values.index(item[Y_KEY])
        heatmap_data[y_idx, x_idx] = item[VALUE_KEY]

    ax = sns.heatmap(
        data=heatmap_data,
        annot=True,
        xticklabels=x_values,
        yticklabels=y_values,
        linewidths=0.5,
    )
    ax.set_xlabel("Batch")
    ax.set_ylabel("Group")
    ax.set_title(title)
    ax.invert_yaxis()
    return ax



from typing import Any, Dict, Literal

from pyspark.sql import DataFrame
from pyspark.sql.functions import col, collect_set
from pyspark.sql.functions import max as spark_max


def ingest_spark_dataframe(
    spark_dataframe: DataFrame,
    save_mode: Literal["Overwrite", "Append"],
    options: Dict[str, Any],
) -> None:
    """
    Saves a Spark DataFrame in multiple batches based on the 'batch' column values.
    Each batch is partitioned by 'group' column and saved according to the provided options.
    Ingest will be performed in parallel according to the `group` column values.

    Parameters
    ----------
    spark_dataframe : DataFrame
        DataFrame containing 'batch' and 'grouping' columns.
        The 'batch' column is used to split the data into separate batches.
        The 'group' column is used for partitioning each batch.
    save_mode : Literal["append", "overwrite"]
        The Spark `mode` to use when ingesting data.
    options : Dict[str, Any]
        Dictionary of options to configure the DataFrame writer.
        Refer to example for more information.

    Example
    -------
    >>> options = {
            "relationship": "BOUGHT",
            "relationship.save.strategy": "keys",
            "relationship.source.save.mode": "Match",
            "relationship.source.labels": ":Customer",
            "relationship.source.node.keys": "customerID:id",
            "relationship.target.save.mode": "Match",
            "relationship.target.labels": ":Product",
            "relationship.target.node.keys": "customerID:id",
            "relationship.properties": "quantity,order"
        }
    >>> save_batches(my_spark_df, "Append", options)

    Note
    ----
    - The input DataFrame must contain 'batch' and 'grouping' columns
    - Each unique value in the 'batch' column will create a separate save operation
    - Uses the Neo4j Spark Connector for writing data to Neo4j
    """

    assert save_mode in {
        "Append",
        "Overwrite",
    }, "`save_mode` must be either 'Append' or 'Overwrite'"
    assert (
        "batch" in spark_dataframe.columns
    ), "Spark DataFrame must contain column `batch`"
    assert (
        "group" in spark_dataframe.columns
    ), "Spark DataFrame must contain column `group`"

    batch_list = spark_dataframe.select(collect_set("batch")).first()[0]

    batches = [
        spark_dataframe.filter(col("batch") == batch_value)
        for batch_value in batch_list
    ]

    num_groups = (
        spark_dataframe.select("group").distinct().count()
    ) 

    # write batches serially to Neo4j database
    for batch in batches:
        (
            batch.repartition(num_groups, "group")  # define parallel groups for ingest
            .write.mode(save_mode)
            .format("org.neo4j.spark.DataSource")
            .options(**options)
            .save()
        )



from pyspark.sql import SparkSession


def verify_spark_version(spark_session: SparkSession) -> None:
    """
    Verify that Spark is a compatible version.
    This package requires Spark >= 3.4.0
    """

    assert spark_session.version >= "3.4.0", "Installed Spark version must be >= 3.4.0"



from pyspark.sql import DataFrame
from pyspark.sql.functions import col


def create_ingest_batches_from_groups(spark_dataframe: DataFrame) -> DataFrame:
    """
    Create batches for ingest into Neo4j.
    Add a `batch` column to the Spark DataFrame identifying which batch the group in that row belongs to.
    Remove `source_group` and `target_group` columns.

    Parameters
    ----------
    spark_dataframe : DataFrame
        The Spark DataFrame to operate on.

    Returns
    -------
    DataFrame
        The Spark DataFrame with `batch` column added.
    """

    # assert that source_group and target_group exist in the dataframe
    # assert that the column types for above are IntegerType()

    source_group_count = spark_dataframe.select("source_group").distinct().count()

    target_group_count = spark_dataframe.select("target_group").distinct().count()

    num_colors = max(source_group_count, target_group_count)

    spark_dataframe = spark_dataframe.withColumn(
        "batch", (col("source_group") + col("target_group")) % num_colors
    ).drop(spark_dataframe.source_group, spark_dataframe.target_group)

    return spark_dataframe



from pyspark.sql import DataFrame
from pyspark.sql.functions import lit


def create_ingest_batches_from_groups(spark_dataframe: DataFrame) -> DataFrame:
    """
    Create batches for ingest into Neo4j.
    Add a `batch` column to the Spark DataFrame identifying which batch the group in that row belongs to.
    In the case of `predefined components` all groups will be in the same batch.

    Parameters
    ----------
    spark_dataframe : DataFrame
        The Spark DataFrame to operate on.

    Returns
    -------
    DataFrame
        The Spark DataFrame with a `batch` column.
    """

    return spark_dataframe.withColumn("batch", lit(0))
